# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
import os, base58, bech32, struct
from binascii import b2a_hex, a2b_hex
from .constants import *
from .compat import hash160, sha256s
from .compat import CT_ecdh, CT_sig_verify, CT_sig_to_pubkey, CT_pick_keypair
from .compat import CT_bip32_derive, CT_priv_to_pubkey

# show bytes as hex in a string
B2A = lambda x: b2a_hex(x).decode('ascii')

def xor_bytes(a, b): # XOR the bytes of A and B
    assert len(a) == len(b)
    return bytes(i^j for i,j in zip(a,b))

def pick_nonce():
    # pick a nonce for our side
    # - must always be a "non trival" value or else card will reject
    for retry in range(3):
        rv = os.urandom(USER_NONCE_SIZE)
        if rv[0] != rv[-1] or len(set(rv)) >= 2:
            return rv

    raise RuntimeError("stuck RNG?")

# Serialization/deserialization tools
# - need this for text msg signing
def ser_compact_size(l):
    if l < 253:
        return struct.pack("B", l)
    elif l < 0x10000:
        return struct.pack("<BH", 253, l)
    elif l < 0x100000000:
        return struct.pack("<BI", 254, l)
    else:
        return struct.pack("<BQ", 255, l)


# high bit set in LE32 indicating hardened BIP-32 path component
HARDENED = 0x8000_0000

def path2str(path):
    # take numeric path (list of numbers) and convert to human form
    # - standardizing on "m/84h" style
    return '/'.join(['m'] + [str(i & ~HARDENED)+('h' if i&HARDENED else '') for i in path])

def str2path(path):
    # normalize notation and return numbers, no error checking
    rv = []

    for i in path.split('/'):
        if i == 'm': continue
        if not i: continue      # trailing or duplicated slashes

        if i[-1] in "p'h":
            here = int(i[:-1], 0) | HARDENED
        else:
            here = int(i, 0)

        rv.append(here)

    return rv

# predicates for numeric paths. stop giggling
all_hardened = lambda path: all(bool(i & HARDENED) for i in path)
none_hardened = lambda path: not any(bool(i & HARDENED) for i in path)


def card_pubkey_to_ident(card_pubkey):
    # convert pubkey into a hash formated for humans
    # - sha256(compressed-pubkey)
    # - skip first 8 bytes of that (because that's revealed in NFC URL)
    # - base32 and take first 20 chars in 4 groups of five
    # - insert dashes
    # - result is 23 chars long
    from base64 import b32encode
    assert len(card_pubkey) == 33, 'expecting compressed pubkey'

    md = b32encode(sha256s(card_pubkey)[8:]).decode('ascii')

    return '-'.join(md[pos:pos+5] for pos in range(0, 20, 5))

def verify_certs(status_resp, check_resp, certs_resp, my_nonce):
    # Verify the certificate chain works, returns label for pubkey recovered from signatures.
    # - raises on any verification issue
    #
    signatures = certs_resp['cert_chain']
    assert len(signatures) >= 2

    r = status_resp
    msg = b'OPENDIME' + r['card_nonce'] + my_nonce
    assert len(msg) == 8 + CARD_NONCE_SIZE + USER_NONCE_SIZE
    pubkey = r['pubkey']

    # check card can sign with indicated key
    ok = CT_sig_verify(pubkey, sha256s(msg), check_resp['auth_sig'])
    if not ok:
        raise RuntimeError("bad sig in verify_certs")

    # follow certificate chain to factory root
    for sig in signatures:
        pubkey = CT_sig_to_pubkey(sha256s(pubkey), sig)

    if pubkey not in FACTORY_ROOT_KEYS:
        # fraudulent device
        raise RuntimeError("Root cert is not from Coinkite. Card is counterfeit.")

    return FACTORY_ROOT_KEYS[pubkey]

def recover_pubkey(status_resp, read_resp, my_nonce, ses_key):
    # [TS] Given the response from "status" and "read" commands,
    # and the nonce we gave for read command, and session key ... reconstruct
    # the card's current pubkey.
    assert status_resp.get('tapsigner', False)

    msg = b'OPENDIME' + status_resp['card_nonce'] + my_nonce + bytes([0])
    assert len(msg) == 8 + CARD_NONCE_SIZE + USER_NONCE_SIZE + 1

    # have to decrypt pubkey
    pubkey = read_resp['pubkey']
    pubkey = pubkey[0:1] + xor_bytes(pubkey[1:], ses_key)

    # Critical: proves card knows key
    ok = CT_sig_verify(pubkey, sha256s(msg), read_resp['sig'])
    if not ok:
        raise RuntimeError("Bad sig in recover_pubkey")

    return pubkey

def recover_address(status_resp, read_resp, my_nonce):
    # [SC] Given the response from "status" and "read" commands, and the
    # nonce we gave for read command, reconstruct the card's verified payment
    # address. Check prefix/suffix match what's expected
    assert not status_resp.get('tapsigner', False)

    sl = status_resp['slots'][0]
    msg = b'OPENDIME' + status_resp['card_nonce'] + my_nonce + bytes([sl])
    assert len(msg) == 8 + CARD_NONCE_SIZE + USER_NONCE_SIZE + 1

    pubkey = read_resp['pubkey']

    # Critical: proves card knows key
    ok = CT_sig_verify(pubkey, sha256s(msg), read_resp['sig'])
    if not ok:
        raise RuntimeError("Bad sig in recover_address")

    expect = status_resp['addr']
    left = expect[0:expect.find('_')]
    right = expect[expect.rfind('_')+1:]

    # Critical: counterfieting check
    addr = render_address(pubkey, status_resp.get('testnet', False))
    if not (addr.startswith(left)
                and addr.endswith(right)
                and len(left) == len(right) == ADDR_TRIM):
        raise RuntimeError("Corrupt response")

    return pubkey, addr

def force_bytes(foo):
    # convert strings to bytes where needed
    return foo.encode('ascii') if isinstance(foo, str) else foo

def calc_xcvc(cmd, card_nonce, his_pubkey, cvc):
    # Calcuate session key and xcvc value need for auth'ed commands
    # - also picks an arbitrary keypair for my side of the ECDH?
    # - requires pubkey from card and proposed CVC value
    assert 6 <= len(cvc) <= 32

    cvc = force_bytes(cvc)

    # fresh new ephemeral key for our side of connection
    my_privkey, my_pubkey = CT_pick_keypair()

    # standard ECDH
    # - result is sha256s(compressed shared point (33 bytes))
    session_key = CT_ecdh(his_pubkey, my_privkey)

    md = sha256s(card_nonce + cmd.encode('ascii'))
    mask = xor_bytes(session_key, md)[0:len(cvc)]
    xcvc = xor_bytes(cvc, mask)

    return session_key, dict(epubkey=my_pubkey, xcvc=xcvc)

def render_address(pubkey, testnet=False):
    # make the text string used as a payment address

    if len(pubkey) == 32:
        # actually a private key, convert
        pubkey = CT_priv_to_pubkey(pubkey)

    HRP = 'bc' if not testnet else 'tb'
    return bech32.encode(HRP, 0, hash160(pubkey))

def render_wif(privkey, bip_178=False, electrum=False, testnet=False):
    # Show the WIF in useful text format (base58)
    # - we are always trying to do bech32/segwit but hard to communicate that
    # - BIP-178 not accepted by community nor Core
    # - electrum adds a prefix for humans (decent idea)
    # - Core 22.0 does not seem to support bech32 import (assumes legacy or does all, IDK)
    assert len(privkey) == 32
    assert (bip_178 or electrum) or not any([bip_178, electrum])

    prefix = bytes([0x80 if not testnet else 0xef])
    suffix = bytes([0x01 if not bip_178 else 0x11])

    rv = base58.b58encode_check(prefix + privkey + suffix).decode('ascii')

    return ('p2wpkh:' + rv) if electrum else rv

def render_descriptor(address=None, privkey=None, bip_178=False, electrum=False, testnet=False):
    # Create a "descriptor" for Core to understand.
    assert address or privkey
    if privkey:
        rv = 'wpkh(%s)' % render_wif(privkey, testnet=testnet)
    else:
        rv = 'addr(%s)' % address

    # TODO: add checksum

    return rv

def verify_master_pubkey(pub, sig, chain_code, my_nonce, card_nonce):
    # using signature response from 'deriv' command, recover the master pubkey
    # for this slot
    msg = b'OPENDIME' + card_nonce + my_nonce + chain_code
    assert len(msg) == 8 + CARD_NONCE_SIZE + USER_NONCE_SIZE + 32

    ok = CT_sig_verify(pub, sha256s(msg), sig)
    if not ok:
        raise RuntimeError("bad sig in verify_master_pubkey")

    return pub

def verify_derive_address(chain_code, master_pub, testnet=False):
    # re-derive the address we should expect
    # - this is "m/0" in BIP-32 nomenclature
    # - accepts master public key (before unseal) or master private key (after)
    pubkey = CT_bip32_derive(chain_code, master_pub, [0])

    return render_address(pubkey, testnet=testnet), pubkey


def make_recoverable_sig(digest, sig, addr=None, expect_pubkey=None, is_testnet=False):
    # The card will only make non-recoverable signatures (64 bytes)
    # but we usually know the address which should be implied by
    # the signature's pubkey, so we can try all values and discover
    # the correct "rec_id" 
    assert len(digest) == 32
    assert len(sig) == 64

    for rec_id in range(4):
        # see BIP-137 for magic value "39"... perhaps not well supported tho
        try:
            rec_sig = bytes([39 + rec_id]) + sig
            pubkey = CT_sig_to_pubkey(digest, rec_sig)
        except ValueError:
            if rec_id >= 2: continue        # because crypto I don't understand
    
        if expect_pubkey:
            if expect_pubkey != pubkey:
                continue
        if addr:
            got = render_address(pubkey, is_testnet)
            if got.endswith(addr):
                return rec_sig
        else:
            return rec_sig

    # failed to recover right pubkey value
    raise ValueError("sig may not be created by that address/pubkey??")


def str_to_int_path(path):
    # convert text  m/34'/33/44 into list of integers

    rv = []
    for i in path.split('/'):
        if i == 'm': continue
        if not i: continue      # trailing or duplicated slashes
        
        if i[-1] in "'phHP":
            assert len(i) >= 2, i
            here = int(i[:-1]) | 0x80000000
        else:
            here = int(i)
            assert 0 <= here < 0x80000000, here
        
        rv.append(here)

    return rv
    
def render_sats_value(c, u):
    # string value for humans: making this hard to parse on purpose
    if not c and not u:
        return '-zero-'
    elif not u:
        return f'{c:,} sats'
    elif c:
        return f'{c:,} sats (+{u:,} soon)'
    else:
        return f'-zero- (+{u:,} soon)'



# EOF



# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# transport.py
#
# Implement the desktop to card connection for our cards, both TAPSIGNER and SATSCARD.
#
#
import sys, os, cbor2
from binascii import b2a_hex, a2b_hex
from hashlib import sha256
from .utils import *
from .constants import *
from .exceptions import CardRuntimeError
from pprint import pformat
from .compat import hash160, sha256s
from .proto import CKTapCard

# Change this to see traffic details
VERBOSE = False

def find_cards():
    #
    # Search all connected card readers, and find all cards that are present.
    #
    # - generator function.
    #
    from smartcard.System import readers as get_readers
    from smartcard.Exceptions import CardConnectionException, NoCardException

    # emulation running on a Unix socket
    sim = CKTapUnixTransport.find_simulator()
    if sim:
        yield CKTapCard(sim)

    readers = get_readers()
    if not readers:
        raise RuntimeError("No USB card readers found. Need at least one.")

    # search for our card
    for r in readers:
        try:
            conn = r.createConnection()
        except:
            continue
        
        try:
            conn.connect()
            atr = conn.getATR()
        except (CardConnectionException, NoCardException):
            #print(f"Empty reader: {r}")
            continue

        if atr == CARD_ATR:
            tr = CKTapNFCTransport(conn)
            yield CKTapCard(tr)
        else:
            print(f"Got ATR: {atr}")

def find_first():
    # operate on the first card we can find
    for c in find_cards():
        return c

    return None

class CKTapTransportABC:
    #
    # Abstract base class. Low level details about talking our protocol.
    #
    is_emulator = False

    def _send_recv(self, msg):
        # take CBOR encoded request, and round-trip the request + response
        raise NotImplementedError

    def get_ATR(self):
        # ATR = Answer To Reset
        raise NotImplementedError

    def close(self):
        # release resources
        pass

    def send(self, cmd, **args):
        # Serialize command, send it as ADPU, get response and decode

        args = dict(args)
        args['cmd'] = cmd
        msg = cbor2.dumps(args)

        if VERBOSE:
            print(f">> {cmd} (%s)" % ', '.join(k+'='+(str(v) if len(str(v)) < 9 else '...')
                                            for k,v in args.items() if k != 'cmd'))

        # Send and wait for reply
        stat_word, resp = self._send_recv(msg)

        try:
            resp = cbor2.loads(resp) if resp else {}
        except:
            #print("Bad CBOR rx'd from card:\n{B2A(resp)}")
            raise RuntimeError('Bad CBOR from card')
            
        if VERBOSE:
            print("<< ", end='')
            if 'error' not in resp:
                print(', '.join(resp.keys()))
            else:
                print(pformat(resp))

        return stat_word, resp

class CKTapNFCTransport(CKTapTransportABC):
    #
    # For talking to a real card over USB to a reader.
    #
    name = 'NFC'

    def __init__(self, card_conn):
        # Check connection they gave us
        # - if you don't have that, use find_cards instead
        atr = card_conn.getATR()
        assert atr == CARD_ATR, "wrong ATR from card"

        self._conn = card_conn

        # Perform "ISO Select" to pick our app
        # - 00 a4 04 00 (APPID)
        # - probably optional
        sw, resp = self._apdu(0x00, 0xa4, APP_ID, p1=4)
        assert sw == SW_OKAY, "ISO app select failed"

    def close(self):
        # release resources
        self._conn.disconnect()
        del self._conn

    def get_ATR(self):
        return self._conn.getATR()

    def _apdu(self, cls, ins, data, p1=0, p2=0):
        # send APDU to card
        lst = [ cls, ins, p1, p2, len(data)] + list(data)
        resp, sw1, sw2 = self._conn.transmit(lst)
        resp = bytes(resp)
        return ((sw1 << 8) | sw2), resp

    def _send_recv(self, msg):
        # send raw bytes (already CBOR encoded) and get response back
        assert len(msg) <= 255, "msg too long"
        return self._apdu(CBOR_CLA, CBOR_INS, msg)

class CKTapUnixTransport(CKTapTransportABC):
    #
    # Emulation running over a Unix socket.
    #
    is_emulator = True
    name = 'EMU'

    @classmethod
    def find_simulator(cls):
        import os
        FN = '/tmp/ecard-pipe'
        if os.path.exists(FN):
            return cls(FN)
        return None

    def __init__(self, pipename):
        import socket
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(pipename)

    def get_ATR(self):
        return CARD_ATR

    def _send_recv(self, msg):
        # send and receive response back
        self.sock.sendall(msg)
        resp = self.sock.recv(4096)

        if not resp:
            # closed socket causes this
            raise RuntimeError("Emu crashed?")

        return 0x9000, resp


# EOF



#
# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#

__all__ = [ 'proto', 'exceptions', 'transport', 'constants', 'utils', 'sweep' ]

from .version import __version__

# find connected cards
from .transport import find_cards, find_first

# base class for working with cards, wants a transport
from .proto import CKTapCard




#
# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# Exceptions
#

class CardRuntimeError(RuntimeError):
    def __init__(self, msg, code, raw_msg):
        self.code = code
        self.raw_msg = msg
        super().__init__(msg)

# EOF



#
# Compatibility wrapper for "coincurve".
#
# nice docs: <https://ofek.dev/coincurve/api/>
#
# - generally using terribile serializations for signatures (DER)
# - docs do not make it clear what serialization is needed
#
from coincurve.ecdsa import deserialize_compact, serialize_compact, der_to_cdata, cdata_to_der
from coincurve import PrivateKey, PublicKey

def CT_sig_verify(pub, msg_digest, sig):
    assert len(sig) == 64
    der = cdata_to_der(deserialize_compact(sig))
    return PublicKey(pub).verify(der, msg_digest, hasher=None)

def CT_sig_to_pubkey(msg_digest, sig):
    assert len(sig) == 65
    rec_id = sig[0]
    # from BIP-137
    if 31 <= rec_id <= 34:
        rec_id -= 31        # P2PKH compressed (most compatible)
    elif 39 <= rec_id <= 42:
        rec_id -= 39        # P2WPKH (most correct for this project)
    else:
        raise ValueError(f'See BIP-137 for recid encoding, saw: {rec_id}')

    sig2 = sig[1:] + bytes([rec_id])
    nxt = PublicKey.from_signature_and_message(sig2, msg_digest, hasher=None)
    return nxt.format()

def CT_ecdh(pubkey, privkey):
    return PrivateKey(privkey).ecdh(pubkey)

def CT_pick_keypair():
    # Choose pub/private pair, return private key (32 bytes) and compressed pubkey
    pk = PrivateKey()
    return pk.secret, PublicKey.from_secret(pk.secret).format()

def CT_sign(privkey, msg_digest, recoverable=False):
    pk = PrivateKey(privkey)
    if recoverable:
        # provides rec_id at end?
        sig = pk.sign_recoverable(msg_digest, hasher=None)
        bip137 = sig[-1] + 31
        return bytes([bip137]) + sig[0:64]
    else:
        der = pk.sign(msg_digest, hasher=None)
        return serialize_compact(der_to_cdata(der))

def CT_priv_to_pubkey(priv):
    pk = PrivateKey(priv)
    return pk.public_key.format()

def CT_bip32_derive(chain_code, master_priv_pub, subkey_path):
    from bip32 import BIP32

    if len(master_priv_pub) == 32:
        # it's actually a private key (from unsealed slot)
        master = BIP32(chaincode=chain_code, privkey=master_priv_pub)
    else:
        # load 'm'
        master = BIP32(chaincode=chain_code, pubkey=master_priv_pub)

    # derive m/0
    _, pubkey = master.get_extended_pubkey_from_path(subkey_path)

    return pubkey

# EOF



#
# Compatibility wrapper for "wallycore".
#
# see <https://wally.readthedocs.io/en/release_0.8.3/crypto/>
#
# - using good serializations already, just very poor docs
#
from os import urandom

from wallycore import ec_sig_verify, ec_sig_to_public_key
from wallycore import ec_public_key_from_private_key
from wallycore import ec_sig_from_bytes as ec_sig_from_digest       # fix misnomer
from wallycore import ecdh as wally_ecdh                            # preference
from wallycore import EC_FLAG_ECDSA, EC_FLAG_RECOVERABLE

# correct API already!
from wallycore import hash160                           # = ripemd160(sha256(x)) => 20 bytes
from wallycore import sha256 as sha256s

def CT_pick_keypair():
    for retry in range(10):
        priv = urandom(32)
        try:
            pub = ec_public_key_from_private_key(priv)
            return priv, pub
        except ValueError:
            continue
    else:
        raise RuntimeError("stuck rng?")

def CT_priv_to_pubkey(priv):
    return ec_public_key_from_private_key(priv)

def CT_sig_verify(pub, msg_digest, sig):
    assert len(sig) == 64
    try:
        ec_sig_verify(pub, msg_digest, EC_FLAG_ECDSA, sig)

        return True
    except ValueError:
        return False

def CT_sig_to_pubkey(msg_digest, sig):
    assert len(sig) == 65
    return bytes(ec_sig_to_public_key(msg_digest, sig))

def CT_ecdh(pubkey, privkey):
    return wally_ecdh(pubkey, privkey)

def CT_sign(privkey, msg_digest, recoverable=False):
    return ec_sig_from_digest(privkey, msg_digest, 
                EC_FLAG_ECDSA | (EC_FLAG_RECOVERABLE if recoverable else 0))

def CT_bip32_derive(chain_code, master_priv_pub, subkey_path):
    from wallycore import bip32_key_init, bip32_key_get_pub_key, bip32_key_from_parent
    from wallycore import BIP32_VER_MAIN_PUBLIC, BIP32_VER_MAIN_PRIVATE
    from wallycore import BIP32_FLAG_SKIP_HASH, BIP32_FLAG_KEY_PUBLIC

    if len(master_priv_pub) == 32:
        m = bip32_key_init(BIP32_VER_MAIN_PRIVATE, 0, 0, chain_code,
            None, master_priv_pub, None, None)
    else:
        m = bip32_key_init(BIP32_VER_MAIN_PUBLIC, 0, 0, chain_code,
            master_priv_pub, None, None, None)

    node = m
    for sk in subkey_path:
        node = bip32_key_from_parent(node, sk, BIP32_FLAG_SKIP_HASH | BIP32_FLAG_KEY_PUBLIC)

    return bip32_key_get_pub_key(node)

# EOF



#
# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# Wrappers for choice of crypto libraries. AKA API Cleanup
#
# My standards:
# - pubkeys: 33 bytes, always compressed
# - private key: 32 bytes
# - signature: 64 bytes or 65 bytes if recoverable
# - no DER, no PEM, no other serializations
# - message digests (for sig/verify) are already digested
# - ECDSA verify returns bool, doesn't raise exception
# - tragically? these all are libsecp256k1 underneath
#

__all__ = [ 'sha256s', 'hash160', 
            'CT_ecdh', 'CT_sig_verify', 'CT_sig_to_pubkey', 'CT_sign',
            'CT_pick_keypair', 'CT_bip32_derive', 'CT_priv_to_pubkey']

# Fall-back code, might be overriden below.
#
def sha256s(msg):
    # single-shot SHA256
    from hashlib import sha256
    return sha256(msg).digest()

def hash160(x):
    # classic bitcoin nested hashes
    from .ripemd import RIPEMD160
    return RIPEMD160(sha256s(x)).digest()

# Other codes must be implemented elsewhere...
#

def CT_pick_keypair():
    # return (priv, pub)
    raise NotImplementedError

def CT_priv_to_pubkey(pk):
    # return compressed pubkey
    raise NotImplementedError

def CT_sig_verify(pub, msg_digest, sig):
    # returns True or False
    assert len(sig) == 64
    raise NotImplementedError

def CT_sig_to_pubkey(msg_digest, sig):
    # returns a pubkey (33 bytes)
    assert len(sig) == 65
    raise NotImplementedError

def CT_ecdh(his_pubkey, my_privkey):
    # returns a 32-byte session key, which is sha256s(compressed point)
    raise NotImplementedError

def CT_sign(privkey, msg_digest, recoverable=False):
    # returns 64-byte sig
    raise NotImplementedError

def CT_bip32_derive(chain_code, master_priv_pub, subkey_path):
    # return pubkey (33 bytes)
    raise NotImplementedError


try:
    # Wally Core <https://wally.readthedocs.io/en/release_0.8.3/crypto/>
    import wallycore

    from .wrap_wally import hash160, sha256s
    from .wrap_wally import CT_ecdh, CT_sig_verify, CT_sig_to_pubkey, CT_sign
    from .wrap_wally import CT_pick_keypair, CT_bip32_derive, CT_priv_to_pubkey

except ImportError:

    try:
        # Coincurve <https://ofek.dev/coincurve/api/>
        import coincurve

        from .wrap_coincurve import CT_ecdh, CT_sig_verify, CT_sig_to_pubkey, CT_sign
        from .wrap_coincurve import CT_pick_keypair, CT_bip32_derive, CT_priv_to_pubkey

    except ImportError:
        raise RuntimeError("need a crypto library")

# EOF



# (c) Copyright 2022 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# proto.py
#
# Implement the higher-level protocol for cards, both TAPSIGNER and SATSCARD.
#
#
import sys, os, cbor2
from binascii import b2a_hex, a2b_hex
from hashlib import sha256
from .utils import *
from .constants import *
from .exceptions import CardRuntimeError
from .compat import hash160, sha256s

class CKTapCard:
    #
    # Protocol/wrapper for cards. Call methods on this instance to get work done.
    #
    # MAYBE: split into TAPSIGNER vs. SATSCARD subclasses and then some methods
    # which aren't appropriate would not exist in the instance. Seems pointless.
    #
    def __init__(self, transport):
        self.tr = transport
        self.first_look()

    def __repr__(self):
        kk = getattr(self, 'card_ident', '???')
        ty = 'TAPSIGNER' if getattr(self, 'is_tapsigner', False) else 'SATSCARD'
        return '<%s %s via %s: %s> ' % (self.__class__.__name__, ty, self.tr.name, kk)

    def close(self):
        # optional? cleanup connection
        self.tr.close()
        del self.tr

    def send(self, cmd, raise_on_error=True, **args):
        # Send a command, get response, but also catch some card state
        # changes and mirror them in our state.
        # - command is a short string, such as "status"
        # - see the protocol spec for arguments here
        stat_word, resp =  self.tr.send(cmd, **args)

        if stat_word != SW_OKAY:
            # Assume error if ANY bad SW value seen; promote for debug purposes
            if 'error' not in resp:
                resp['error'] = "Got error SW value: 0x%04x" % stat_word
            resp['stat_word'] = stat_word


        if 'card_nonce' in resp:
            # many responses provide an updated card_nonce needed for
            # the *next* comand. Track it.
            # - only changes when "consumed" by commands that need CVC
            self.card_nonce = resp['card_nonce']

        if raise_on_error and 'error' in resp:
            msg = resp.pop('error')
            code = resp.pop('code', 500)
            raise CardRuntimeError(f'{code} on {cmd}: {msg}', code, msg)

        return resp

    def first_look(self):
        # Call this at end of __init__ to load up details from card
        # - can be called multiple times

        st = self.send('status')
        assert 'error' not in st, 'Early failure: ' + repr(st)
        assert st['proto'] == 1, "Unknown card protocol version"
        if st.get('tampered'):
            print("WARNING: Card has set tampered flag!")

        self.card_pubkey = st['pubkey']
        self.card_ident = card_pubkey_to_ident(self.card_pubkey)

        self.applet_version = st['ver']
        self.birth_height = st.get('birth', None)
        self.is_testnet = st.get('testnet', False)
        self.auth_delay = st.get('auth_delay', 0)
        self.is_tapsigner =  st.get('tapsigner', False)
        self.active_slot, self.num_slots = st.get('slots', (0,1))
        assert self.card_nonce      # self.send() will have captured from first status req

        # certs will not verify on emulator, and expensive to do more than once in
        # normal cases too
        self._certs_checked = bool(self.tr.is_emulator)

    def send_auth(self, cmd, cvc, **args):
        # Take CVC and do ECDH crypto and provide the CVC in encrypted form
        # - returns session key and usual auth arguments needed
        # - skip if CVC is None and just do normal stuff (optional auth on some cmds)
        # - for commands w/ encrypted arguments, you must provide to this function

        if cvc:
            session_key, auth_args = calc_xcvc(cmd, self.card_nonce, self.card_pubkey, cvc)
            args.update(auth_args)
        else:
            session_key = None

        # A few commands take an encrypted argument (most are returning encrypted
        # results) and the caller didn't know the session key yet. So xor it for them.
        if cmd == 'sign':
            args['digest'] = xor_bytes(args['digest'], session_key)
        elif cmd == 'change':
            args['data'] = xor_bytes(args['data'], session_key[0:len(args['data'])])

        return session_key, self.send(cmd, **args)


    #
    # Wrappers and Helpers
    #
    def address(self, faster=False, incl_pubkey=False, slot=None):
        # Get current payment address for card
        # - does 100% full verification by default
        # - returns a bech32 address as a string
        assert not self.is_tapsigner

        # check certificate chain
        if not self._certs_checked and not faster:
            self.certificate_check()

        st = self.send('status')
        cur_slot = st['slots'][0]
        if slot is None:
            slot = cur_slot

        if ('addr' not in st) and (cur_slot == slot):
            #raise ValueError("Current slot is not yet setup.")

            return None

        if slot != cur_slot:
            # Use the unauthenticated "dump" command.
            rr = self.send('dump', slot=slot)

            assert not incl_pubkey, 'can only get pubkey on current slot'

            return rr['addr']
        
        # Use special-purpose "read" command
        n = pick_nonce()
        rr = self.send('read', nonce=n)
        
        pubkey, addr = recover_address(st, rr, n)

        if not faster:
            # additional check: did card include chain_code in generated private key?
            my_nonce = pick_nonce()
            card_nonce = self.card_nonce
            rr = self.send('derive', nonce=my_nonce)
            master_pub = verify_master_pubkey(rr['master_pubkey'], rr['sig'],
                                                rr['chain_code'], my_nonce, card_nonce)
            derived_addr,_ = verify_derive_address(rr['chain_code'], master_pub,
                                                        testnet=self.is_testnet)
            if derived_addr != addr:
                raise ValueError("card did not derive address as expected")

        if incl_pubkey:
            return pubkey, addr

        return addr

    def get_derivation(self):
        # TAPSIGNER only: what's the current derivation path, which might be
        # just empty (aka 'm').
        assert self.is_tapsigner
        st = self.send('status')
        path = st.get('path', None)
        if path is None:
            #raise RuntimeError("no private key picked yet, so no derivation")
            return None
        return path2str(path)

    def set_derivation(self, path, cvc):
        # TAPSIGNER only: what's the current derivation path, which might be
        # just empty (aka 'm').
        assert self.is_tapsigner
        np = str2path(path)

        if not all_hardened(np):
            raise ValueError("All path components must be hardened")

        _, resp = self.send_auth('derive', cvc, path=np, nonce=pick_nonce())

        # XXX need FP of parent key and master (XFP)
        # XPUB would be better result here, but caller can use get_xpub() next

        return len(np), resp['chain_code'], resp['pubkey']

    def get_xfp(self, cvc):
        # fetch master xpub, take pubkey from that and calc XFP
        assert self.is_tapsigner
        _, st = self.send_auth('xpub', cvc, master=True)
        xpub = st['xpub']
        return hash160(xpub[-33:])[0:4]

    def get_xpub(self, cvc, master=False):
        # fetch XPUB, either derived or master one
        # - result is BIP-32 serialized and base58-check encoded
        assert self.is_tapsigner
        _, st = self.send_auth('xpub', cvc, master=master)
        xpub = st['xpub']
        return base58.b58encode_check(xpub).decode('ascii')

    def get_pubkey(self, cvc):
        # TAPSIGNER only: Get the public key for current derived path
        # - on TS, it's an authenticated command: 'read'
        # - equiv. to get_xpub(master=False) and looking at part of that value
        # - fairly pointless
        assert self.is_tapsigner
        n = pick_nonce()
        st = self.send('status')
        ses_key, rr = self.send_auth('read', cvc, nonce=n)

        return recover_pubkey(st, rr, n, ses_key)

    def make_backup(self, cvc):
        # read the backup file; gives ~100 bytes to be kept long term
        assert self.is_tapsigner
        _, st = self.send_auth('backup', cvc)
        return st['data']

    def change_cvc(self, old_cvc, new_cvc):
        # Change CVC. Note: can be binary or ascii or digits, 6..32 long
        assert 6 <= len(new_cvc) <= 32
        _, st = self.send_auth('change', old_cvc, data=force_bytes(new_cvc))

    def certificate_check(self):
        # Verify the certificate chain and the public key of the card
        # - assures this card was produced in Coinkite factory
        # - does not relate to payment addresses or slot usage
        # - raises on errors/failed validation
        st = self.send('status')
        certs = self.send('certs')

        n = pick_nonce()
        check = self.send('check', nonce=n)

        rv = verify_certs(st, check, certs, n)
        self._certs_checked = True

        return rv

    def get_status(self):
        # read current status
        return self.send('status')

    def unseal_slot(self, cvc):
        # Unseal the current slot (can only be one)
        # - returns (privkey, slot_num)
        assert not self.is_tapsigner

        # only one possible value for slot number
        target = self.active_slot

        # but that slot must be used and sealed (note: unauthed req here)
        resp = self.send('dump', slot=target)

        if resp.get('sealed', None) == False:
            raise RuntimeError(f"Slot {target} has already been unsealed.")

        if resp.get('sealed', None) != True:
            raise RuntimeError(f"Slot {target} has not been used yet.")

        ses_key, resp = self.send_auth('unseal', cvc, slot=target)

        pk = xor_bytes(ses_key, resp['privkey'])

        return pk, target

    def get_nfc_url(self):
        # Provide the (dynamic) URL that you'd get if you tapped the card.
        return self.send('nfc').get('url')

    def get_privkey(self, cvc, slot):
        # Provide the private key of an already-unsealed slot (32 bytes)
        assert not self.is_tapsigner
        ses_key, resp = self.send_auth('dump', cvc, slot=slot)

        if 'privkey' not in resp:
            if resp.get('used', None) == False:
                raise RuntimeError(f"That slot ({slot}) is not yet used (no key yet)")
            if resp.get('sealed', None) == True:
                raise RuntimeError(f"That slot ({slot}) is not yet unsealed.")

            # unreachable?
            raise RuntimeError(f"Not sure of the key for that slot ({slot}).")

        return xor_bytes(ses_key, resp['privkey'])

    def get_slot_usage(self, slot, cvc=None):
        # Get address and status for a slot, CVC is optional
        # returns:
        #   (addr, status, detail_map) 
        assert not self.is_tapsigner
        session_key, here = self.send_auth('dump', cvc, slot=slot)

        addr = here.get('addr', None)
        if here.get('sealed', None) == True:
            status = 'sealed'
            if slot == self.active_slot:
                addr = self.address(faster=True)
        elif (here.get('sealed', None) == False) or ('privkey' in here):
            status = 'UNSEALED'
            if 'privkey' in here:
                pk = xor_bytes(session_key, here['privkey'])
                addr = render_address(pk, self.is_testnet)
        elif here.get('used', None) == False:
            status = "unused"
        else:
            # unreachable.
            raise ValueError(repr(here))

        addr = addr or here.get('addr')

        return (addr, status, here)

    # TODO
    # - 'sign_digest' command which does the retries needed
    # - 'wait' command which does delay needed, if any (no UX)

# EOF



#
# (c) Copyright 2022 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
import os
from .compat import sha256s
from .compat import CT_sig_to_pubkey
from .utils import card_pubkey_to_ident, render_address


def all_keys(sig, md):
    # generates all possible pubkeys from sig + digest
    for rec_id in range(4):
        # see BIP-137 for magic value "39"... perhaps not well supported tho
        try:
            yield CT_sig_to_pubkey(md, bytes([39 + rec_id]) + sig)
        except ValueError:
            if rec_id >= 2: continue        # because crypto I don't understand
            raise

def url_decoder(fragment):
    # Takes the URL (after the # part) and verifies it
    # and returns dict of useful values, or raise on errors/frauds
    from urllib.parse import parse_qsl

    assert '#' not in fragment
    assert '?' not in fragment

    msg = fragment[0:fragment.rfind('=')+1]
    try:
        raw = dict(parse_qsl(fragment, strict_parsing=True))
    except:
        raise RuntimeError("Badly formated link")

    try:
        nonce = bytes.fromhex(raw['n'])
        is_tapsigner = bool(raw.get('t', False))
        assert len(nonce) == 8
        slot_num = int(raw.get('o', -1))
        addr = raw.get('r', None)
        sig = bytes.fromhex(raw['s'])
        assert len(sig) == 64
        card_ident = raw.get('c', None)
    except KeyError as exc:
        raise RuntimeError("Required field missing")

    md = sha256s(msg.encode('ascii'))

    if is_tapsigner:
        assert card_ident, 'missing card ident value'
        card_ident = bytes.fromhex(card_ident)
        full_card_ident = None

        for pubkey in all_keys(sig, md):
            expect = sha256s(pubkey)
            if expect[0:8] == card_ident:
                full_card_ident = card_pubkey_to_ident(pubkey)
                break

        if not full_card_ident:
            raise RuntimeError("Could not reconstruct card ident.")

        return dict(nonce=nonce.hex(),
                    card_ident=full_card_ident,
                    virgin=(raw['u'] == 'U'),
                    is_tapsigner=True, tampered=(raw['u'] == 'E'))

    else:
        # SATSCARD
        confirmed_addr = None
        is_testnet = False
        state = dict(S='Sealed', U='UNSEALED', E='Error/Tampered').get(raw['u'], 'Unknown state')

        for pubkey in all_keys(sig, md):
            if addr is not None:
                got = render_address(pubkey, False)
                if got.endswith(addr):
                    confirmed_addr = got
                    break

                got = render_address(pubkey, True)
                if got.endswith(addr):
                    confirmed_addr = got
                    is_testnet = True
                    break

        if addr and not confirmed_addr:
            raise RuntimeError("Could not reconstruct full payment address.")

        rv = dict(state=state, addr=confirmed_addr, nonce=nonce.hex(),
                    is_tapsigner=False,
                    slot_num=slot_num,
                    sealed=(raw['u'] == 'S'),
                    tampered=(raw['u'] == 'E'))

        if is_testnet:
            rv['testnet'] = True

        return rv





#
# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# System constants.
#

# Number of key slots in SATSCARD
NUM_SLOTS = 10

# the "CVC" is the spending code on back of card.
# - see also many text messages to user about this
# - for TAPSIGNER, this is a minimum length (max 32)
CVC_LENGTH = 6

# no need to scan the blockchain before this point in time, since product didn't exist yet
# - March 25/2022
PROJECT_EPOC_TIME_T = 1648215566

# length from start/end of bech32 address that is provided
# - center part will be replaced with three underscores
ADDR_TRIM = 12

# require nonce sizes (bytes)
CARD_NONCE_SIZE = 16
USER_NONCE_SIZE = 16

# published Coinkite factory root keys
FACTORY_ROOT_KEYS = { 
    bytes.fromhex('03028a0e89e70d0ec0d932053a89ab1da7d9182bdc6d2f03e706ee99517d05d9e1'):
        'Root Factory Certificate',

    # obsolete dev value, but keeping for a little while longer
    bytes.fromhex('027722ef208e681bac05f1b4b3cc478d6bf353ac9a09ff0c843430138f65c27bab'):
        'Root Factory Certificate (TESTING ONLY)',
}

# our cards will provide this answer to reset (ATR)
CARD_ATR = [59, 136, 128, 1] + list(b'Coinkite') + [49]

# our Javacard applet has this APP ID
APP_ID = bytes.fromhex('f0436f696e6b697465434152447631')

# APDU CLA and INS fields for our one APDU, which uses CBOR data
CBOR_CLA = 0x00
CBOR_INS = 0xCB

# Correct ADPU response from all commands: 90 00 
SW_OKAY = 0x9000

# EOF



#
# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# Functions related to sweeping funds and building transactions.
#
# - Requires 'requests[socks]' module
# - Will try to use Tor if you have it running locally already
# - Uses data from <blockstream.info> by default
#
import sys, os, time, json
from pprint import pformat
from getpass import getpass
from collections import namedtuple
from cktap.utils import render_sats_value

# Explora protocol <https://github.com/Blockstream/esplora/blob/master/API.md>
DEFAULT_SERVER = 'https://blockstream.info'
ONION_SERVER = 'http://explorerzydxu5ecjrkwceayqybizmpjjznk5izmitf2modhcusuqlid.onion'

# we check for any of these ports being open and assume it's Tor if found
LOCALHOST_PROXY_PORTS = [ 9150, 9050 ]

class NetConnection:

    def __init__(self, server=None):
        import requests
        self.ses = requests.Session()
        self.is_tor = self.tor_upgrade()
        self.server = server or (DEFAULT_SERVER if not self.is_tor else ONION_SERVER)
        assert not self.server.endswith('/')

        # I want no user-agent header at all, so have to
        # use this one strange hack on urllib3...
        from urllib3.util import SKIP_HEADER
        self.ses.headers['user-agent'] = SKIP_HEADER

    def tor_upgrade(self):
        # See if Tor is running, if so, apply socks-proxy details
        # - not doing generalized socks proxies but
        # - you can override with HTTP_PROXY / HTTPS_PROXY in environment
        #   which would be directly implemented in requests, see
        #   <https://2.python-requests.org/en/master/user/advanced/#proxies>
        import requests

        if ('HTTP_PROXY' in os.environ) or ('HTTPS_PROXY' in os.environ) or self.ses.proxies:
            return False

        try:
            # do we have support for socks?
            import socks
        except ImportError:
            return False

        for port in LOCALHOST_PROXY_PORTS:
            try:
                r = requests.get(f'http://127.0.0.1:{port}')
            except:
                continue
            if ('This is a SOCKS Proxy, Not An HTTP Proxy' in r.text):
                # usual socks error message from Tord running locally.
                self.ses.proxies['https'] = f'socks5h://127.0.0.1:{port}'
                self.ses.proxies['http'] = f'socks5h://127.0.0.1:{port}'

                return True

        return False
            
    def get_json(self, path, **kws):
        # fetch a JSON response
        assert path[0] == '/'
        r = self.ses.get(self.server + path, **kws)
        r.raise_for_status()
        try:
            return r.json()
        except json.decoder.JSONDecodeError:
            raise ValueError("Bad json: " + r.text)
            


    
UTXO = namedtuple('UTXO', 'txid vout value height confirmed')

class UTXOList:

    def __init__(self, address, slot_num=None, server=None):
        # must call self.fetch() after setup
        self.slot = slot_num
        self.addr = address
        self.testnet = address.startswith('tb1')
        self.web = NetConnection(server)
        self.utxos = []


    def fetch(self):
        # load up the data from network
        # - TODO: cursor support for > 25 utxo? Not clear if supported on server

        path = ('/testnet' if self.testnet else '') + f'/api/address/{self.addr}/utxo'
        ans = self.web.get_json(path)
        for u in ans:
            h = u['status'].get('block_height', -1)
            conf = u['status'].get('confirmed', False)

            utxo = UTXO(u['txid'], u['vout'], u['value'], h, conf)

            self.utxos.append(utxo)

        return len(self.utxos)

    # never, ever, add these values together!
    def confirmed_balance(self):
        return sum(u.value for u in self.utxos if u.confirmed)
    def unconfirmed_balance(self):
        return sum(u.value for u in self.utxos if not u.confirmed)

    def balance(self):
        # string value for humans, start with this
        c = self.confirmed_balance()
        u = self.unconfirmed_balance()
        return render_sats_value(c, u)

    def fetch_txns(self):
        # Fetch transaction details for all desposit transactions
        # - should be all the data needed to build a transaction for signing
        # - except for miner fee, and reconstructing redeem scripts, etc...
        # - oh, and hashing correctly
        prefix = '/testnet' if self.testnet else ''

        txns = {}
        for txid in set(u.txid for u in self.utxos):
            txns[txid] = self.web.get_json(prefix + f'/api/tx/{txid}')

        return txns

# EOF



#
# Coinkite Tap protocol and python support library
#

# To use this command to install and yet be able to edit the code (here). Great for dev:
#
#   pip install --editable .
#
# On Windows, this can be useful:
#
#   pip install -U --force-reinstall coinkite_tap_protocol-*.whl[cli]
#
#
from setuptools import setup

# these minimum versions are tested, some earlier values would probably work too.
requirements = [
    'cbor2>=5.4.1',
    'bech32>=1.2.0',
    'base58>=2.1.1',
    'pyscard>=2.0.2',
]

if 1:
    # We support wallycore or coincurve; both of which
    # ultimately call libsecp256k1 
    #
    # - but I could not compile wallycore on windows
    # - prolly because coincurve has 500+ lines of setup.py code
    # - sorry I could not make this file detect O/S automatically?!
    #
    requirements.extend([
        'coincurve>=15.0.1',
        'bip32>=2.1',
    ])
else:
    requirements.extend([
        'wallycore>=0.8.2',
    ])

cli_requirements = [
    'click>=8.0.3',
    'pyqrcode>=1.2.1',
    'pypng>=0.0.21',
    'requests[socks]>=2.26.0',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

from cktap.version import __version__

setup(
    name='coinkite-tap-protocol',
    version=__version__,
    packages=[ 'cktap' ],
    python_requires='>3.6.0',
    install_requires=requirements,
    extras_require={
        'cli': cli_requirements,
    },
    url='https://github.com/coinkite/coinkite-tap-proto',
    author='Coinkite Inc.',
    author_email='support@coinkite.com',
    description="Communicate with your TAPSIGNER or SATSCARD using Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points='''
        [console_scripts]
        cktap=cktap.cli:main
    ''',
    classifiers=[
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
    ],
)




#!/usr/bin/env python
#
# (c) Copyright 2021 by Coinkite Inc. This file is covered by license found in COPYING-CC.
#
# To use this, install with:
#
#   pip install --editable .
#
# That will create the command "cktap" in your path.
#
#
import click, sys, os, pdb, time, json, datetime
from pprint import pformat
from binascii import b2a_hex, a2b_hex
from functools import wraps
from getpass import getpass

from .utils import xor_bytes, render_address, render_wif, render_descriptor, B2A, ser_compact_size
from .utils import make_recoverable_sig, render_sats_value, path2str, pick_nonce
from .compat import sha256s
from .constants import *
from .exceptions import CardRuntimeError
from .transport import CKTapCard, find_cards

# dict of options that apply to all commands
global global_opts
global_opts = dict()

# Cleanup display (supress traceback) for user-feedback exceptions
_sys_excepthook = sys.excepthook
def my_hook(ty, val, tb):
    if ty in { CardRuntimeError, RuntimeError }:
        print("\n\nFATAL: %s" % val, file=sys.stderr)
    else:
        return _sys_excepthook(ty, val, tb)
sys.excepthook=my_hook

def fail(msg):
    # show message and stop
    click.echo(msg)
    sys.exit(1)

def get_card(only_satscard=False, only_tapsigner=False):
    # Pick a card to work with
    global global_opts
    ci_filter = (global_opts.get('card_ident') or '').upper()
    wait_for_it = global_opts.get('wait', False)

    be_verbose = global_opts.get('verbose', False)
    if be_verbose:
        import cktap.transport as tt
        tt.VERBOSE = True

    first = True
    while 1:
        for c in find_cards():
            if ci_filter:
                if not (ci_filter in c.card_ident):
                    c.close()
                    continue
            if only_satscard and c.is_tapsigner: continue
            if only_tapsigner and not c.is_tapsigner: continue
            return c

        if not wait_for_it: 
            subset = 'matching cards' if ci_filter else 'suitable cards'
            if only_tapsigner: subset = 'TAPSIGNER cards' 
            if only_satscard: subset = 'SATSCARD'
            fail(f"No {subset} found. Is it in place on reader?")

        if first:
            click.echo("Waiting for card...")
            first = False

        time.sleep(1)
        

def dump_dict(d):

    for k,v in d.items():
        if k == 'card_nonce':
            # no point showing this value
            continue

        if isinstance(v, (bytes, bytearray)):
            v = B2A(v)

        click.echo('%s: %s' % (k, v))

def cleanup_cvc(card, cvc, missing_ok=False, prompt="spending code", confirm=False):
    # Cleanup CVC provided (digits only) and prompt if needed, fail if wrong length
    if not cvc:
        if missing_ok:
            return None
        cvc = getpass(f"Enter {prompt}: ")

        if confirm:
            chk = getpass(f"Repeat {prompt}: ")
            if chk != cvc:
                fail("Does not match first try!? Stop.")

    if not card.is_tapsigner:
        # remove non-digits
        cvc = cvc[0:0].join(d for d in cvc if d.isdigit())

    if not cvc and missing_ok:
        # blank/empty response, ok in this case
        return None

    if card.is_tapsigner:
        if len(cvc) < 6:
            fail(f"CVC is too short, must be at least 6 chars.")
        if len(cvc) > 32:
            fail(f"CVC is too long, must be at most 32 chars.")
    else:
        if len(cvc) != CVC_LENGTH:
            fail(f"Need {CVC_LENGTH}-digit numeric code from back of card.")

    if card.auth_delay:
        with click.progressbar(label="Requires security delay", length=card.auth_delay) as bar:
            for n in range(card.auth_delay):
                card.send('wait')
                bar.update(1)

    return cvc
    
def display_errors(f):
    # clean-up display of errors from device
    # XXX use me
    @wraps(f)
    def wrapper(*args, **kws):
        try:
            return f(*args, **kws)
        except RuntimeError as exc:
            click.echo("\n%s\n" % str(exc.args[0]))
            sys.exit(1)
    return wrapper

# Accept any prefix of a command name.
#
# from <https://click.palletsprojects.com/en/8.0.x/advanced/?#command-aliases>
class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Abiguous command. Pick one of: {' | '.join(sorted(matches))}")

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


#
# Options we want for all commands
#
@click.group(cls=AliasedGroup)
@click.option('--card-ident', '-i', default=None, metavar="BLAHZ-",
                    help="Operate on specific card (any substring is enough)")
@click.option('--wait', '-w', is_flag=True, 
                    help="Waits until a card is in place.")
@click.option('--verbose', '-v', is_flag=True, 
                    help="Show traffic with card.")
@click.option('--pdb', is_flag=True, 
                    help="Prepare patient for surgery to remove bugs.")
def main(**kws):
    '''
    Interact with SATSCARD and TAPSIGNER cards via NFC tap.

    Command marked [TS] are only for TAPSIGNER and [SC] only for SATSCARD.

    You can use "bal", or "b" for "balance": any distinct prefix for all commands.

    '''

    # implement PDB option here
    if kws.pop('pdb', False):
        import pdb, sys
        def doit(ex_cls, ex, tb):
            pdb.pm()
        sys.excepthook = doit

    # global options, mostly not considered here
    global global_opts
    global_opts.update(kws)
        
@main.command('debug')
def interactive_debug():
    "Start interactive (local) debug session."
    import code
    import atexit
    import os
            
    try:
        import readline
        class HistoryConsole(code.InteractiveConsole):
            def __init__(self, locals=None, filename="<console>",
                         histfile=os.path.expanduser("~/.console-history")):
                code.InteractiveConsole.__init__(self, locals, filename)
                self.init_history(histfile)
            
            def init_history(self, histfile):
                readline.parse_and_bind("tab: complete")
                if hasattr(readline, "read_history_file"):
                    try:
                        readline.read_history_file(histfile)
                    except IOError:
                        pass
                    atexit.register(self.save_history, histfile)
            
            def save_history(self, histfile):
                readline.write_history_file(histfile)
    except ImportError:
        # windows probably, keep going
        HistoryConsole = code.InteractiveConsole

    # useful stuff
    import pdb
    from pdb import pm
    C = get_card()
    S = C.send
    SA = C.send_auth

    cli = HistoryConsole(locals=dict(globals(), **locals()))
    cli.interact(banner="""\
Go for it: 'C' is the connected card, S=C.send ... S('status') SA('cmd', cvc, ...)""", exitmsg='')

@main.command('chain')
def get_block_chain():
    '''[SC] Get which blockchain (Bitcoin/Testnet) is configured.

    BTC=>Bitcoin  or  XTN=>Bitcoin Testnet
    '''

    card = get_card(only_satscard=True)

    click.echo('XTN' if card.is_testnet else 'BTC')

    
@main.command('msg')
@click.argument('message')
@click.argument('cvc', type=str, metavar="(6-digit # code)", required=False)
@click.option('--verbose', '-v', is_flag=True, help='[SC] Include full ascii armour')
@click.option('--just-sig', '-j', is_flag=True, help='Just the signature itself, nothing more')
@click.option('--slot', '-s', type=int, metavar="#", default=0, help="Slot number, default: zero")
def sign_message(cvc, message, path=2, verbose=True, just_sig=False, slot=0):
    "Sign a short text message (TODO -- INCOMPLETE)"
    from base64 import b64encode

    card = get_card()
    cvc = cleanup_cvc(card, cvc)

    message = message.encode('ascii') if not isinstance(message, bytes) else message

    # TODO: 
    # - using <https://github.com/bitcoin/bips/blob/master/bip-0322.mediawiki>
    # - build a message digest, based on BIP-340 "tagged hash" and a fake to_sign txn
    # - send digest to card
    # - serialize result, which includes to_sign txn

    # XXX until then, pretend we are living in a simple 2010 world.
    # - I don't know of any tools which can be used to verify this signature... so it's useless
    # XXX on TS, this could work and be useful .. because we'd just share a classic address
    xmsg = b'\x18Bitcoin Signed Message:\n' + ser_compact_size(len(message)) + message
    md = sha256s(sha256s(xmsg))

    ses_key, resp = card.send_auth('sign', cvc, slot=slot, digest=md)
    expect_pub = resp['pubkey']

    if card.is_tapsigner:
        addr = None
        just_sig = True
    else:
        addr = card.address(slot=slot)

    # problem: not a recoverable signature, need to calc recid based on our
    # knowledge of address
    raw = make_recoverable_sig(md, resp['sig'],
                                addr=addr, expect_pubkey=expect_pub,
                                is_testnet=card.is_testnet)

    sig = str(b64encode(raw), 'ascii').replace('\n', '')

    if just_sig:
        click.echo(str(sig))
    else:
        if verbose:
            click.echo('-----BEGIN SIGNED MESSAGE-----\n{msg}\n-----BEGIN '
                      'SIGNATURE-----\n{addr}\n{sig}\n-----END SIGNED MESSAGE-----'.format(
                            msg=message.decode('ascii'), addr=addr, sig=sig))
        else:
            click.echo('%s\n%s\n%s' % (message.decode('ascii'), addr, sig))
    
@main.command('version')
def get_version():
    "Get the version of the card's firmware installed (but not upgradable)"

    card = get_card()

    click.echo(card.applet_version)

@main.command('list')
def list_cards():
    "List all cards detected on any reader attached."

    count = 0
    for card in find_cards():
        click.echo(repr(card))
        count += 1

    if not count:
        click.echo("(none found)")

@main.command('usage')
@click.argument('cvc', type=str, metavar="(6-digit # code)", required=False)
def get_usage(cvc):
    "[SC] Show slots usage so far."

    card = get_card(only_satscard=True)
    cvc = cleanup_cvc(card, cvc, missing_ok=True)

    print('SLOT# |  STATUS  | ADDRESS')
    print('------+----------+-------------')
    for slot in range(card.num_slots):
        addr, status, _ = card.get_slot_usage(slot, cvc=cvc)
        
        print('%3d   | %-8s | %s' % (slot, status, addr or ''))

@main.command('address')
def get_addr():
    "[SC] Show current deposit address"
    card = get_card(only_satscard=True)

    addr = card.address()
    if not addr:
        fail("Current slot not yet setup and has no address.")

    click.echo(addr)

@main.command('open')
@click.option('--slot', '-s', type=int, metavar="#", default=None, help="Slot number (optional)")
def get_addr_open_app(slot):
    "[SC] Get address and open associated local Bitcoin app to handle it"
    card = get_card(only_satscard=True)

    addr = card.address(slot=slot)
    if not addr:
        fail("Current slot not yet setup and has no address.")

    url = 'bitcoin:' + addr

    click.launch(url)

@main.command('url')
@click.option('--open-browser', '-o', is_flag=True, 
                help="Launch web browser app to view the URL")
def get_nfc_url(open_browser):
    "Get website URL used for NFC verification, and optionally open it"
    card = get_card()

    url = card.get_nfc_url()

    click.echo(url)

    if open_browser:
        click.launch(url)

@main.command('qr')
@click.option('--outfile', '-o', metavar="filename.png",
                        help="Save an SVG or PNG (depends on extension)", default=None,
                        type=click.File('wb'))
@click.option('--slot', '-s', type=int, metavar="#", default=None, help="Slot number (optional)")
@click.option('--error-mode', '-e', default='L', metavar="L|M|H",
            help="Forward error correction level (L = low, H=High=bigger)")
def get_deposit_qr(outfile, slot, error_mode):
    "[SC] Show current deposit address as a QR (or private key if unsealed)"
    import pyqrcode

    card = get_card(only_satscard=True)

    addr = card.address(slot=slot)
    if not addr:
        fail("Current slot not yet setup and has no address.")

    url = 'bitcoin:{addr}'.format(addr=addr).upper()
    q = pyqrcode.create(url, error=error_mode, mode='alphanumeric')

    if not outfile:
        # TODO: this doesn't work on Windows
        print(q.terminal(quiet_zone=2))
        print((' '*12) + addr)
        print()
    else:
        if outfile.name.lower().endswith('.svg'):
            q.svg(outfile, scale=1)
        else:
            q.png(outfile)

        click.echo(f"Wrote {outfile.tell():,} bytes to: {outfile.name}", err=1)

@main.command('dump')
@click.argument('slot', type=int, metavar="[SLOT#]", required=False, default=0)
@click.argument('cvc', type=str, metavar="[6-digit code]", required=False)
def dump_slot(slot, cvc):
    "[SC] Show state of slot number indicated. Needs CVC to get more info on unsealed slots."
    # XXX too low-level/hexy
    card = get_card(only_satscard=True)

    cvc = cleanup_cvc(card, cvc, missing_ok=True)
    session_key, resp = card.send_auth('dump', cvc, slot=slot)
    if 'privkey' in resp:
        resp['privkey'] = xor_bytes(session_key, resp['privkey'])

    dump_dict(resp)
        
@main.command('check')
@click.argument('cvc', type=str, metavar="(6-digit # code)", required=False)
def check_cvc(cvc):
    "Verify you have the spending code (CVC) correct. Does nothing with it"

    card = get_card()
    cvc = cleanup_cvc(card, cvc)

    # do a dump command
    if card.is_tapsigner:
        # a command w/o side effects for TS
        cmd = 'read'
        args = dict(nonce=pick_nonce())
    else:
        cmd = 'dump'
        args = dict(slot=0)

    try:
        ses_key, resp = card.send_auth(cmd, cvc, **args)
        click.echo("Code is correct.")
    except CardRuntimeError as exc:
        if exc.code == 429:
            # not expected
            fail("Rate limited. Need to wait.")
        elif exc.code == 401:
            fail("Incorrect code.")
        else:
            raise

@main.command('path')
def get_path():
    "[TS] Show the subkey derivation path in effect"
    card = get_card(only_tapsigner=True)

    path = card.get_derivation()
    if path == None:
        fail("NONE (no key picked yet)")
    else:
        print(path)
    
@main.command('derive')
@click.argument('path', type=str, metavar="84h/0h/0h", required=True)
@click.argument('cvc', type=str, metavar="[6-digit code]", required=False)
@click.option('--skip-checks', '-x', is_flag=True, help="Skip extra checking of pubkey")
def set_derivation(path, cvc, skip_checks):
    "[TS] Change the subkey derivation path to use (shows xpub)"
    card = get_card(only_tapsigner=True)

    cvc = cleanup_cvc(card, cvc)
    child_depth, cc, pub = card.set_derivation(path, cvc)

    xp = card.get_xpub(cvc)

    if not skip_checks:
        # extra checking; robust against MitM (cost = 20ms)
        from base58 import b58decode_check
        expect_pubkey = b58decode_check(xp)[-33:]

        got_pubkey = card.get_pubkey(cvc)
        assert expect_pubkey == got_pubkey

    print(xp)
    
        
@main.command('certs')
def check_certs():
    "Check this card was made by Coinkite: Verifies a certificate chain up to root factory key."

    card = get_card()

    label = card.certificate_check()
    
    click.echo("Genuine card from Coinkite.\n\nHas cert signed by: %s" % label)
    
@main.command('unseal')
@click.argument('cvc', type=str, metavar="(6-digit # code)", required=False)
def unseal_slot(cvc):
    "[SC] Unseal current slot and reveal private key. Does not setup next slot."

    card = get_card(only_satscard=True)
    cvc = cleanup_cvc(card, cvc)

    pk, slot_num = card.unseal_slot(cvc)

    # show all the details
    dump_key_info(slot_num, pk, is_testnet=card.is_testnet)

@main.command('wif')
@click.option('--slot', '-s', type=int, metavar="#", default=-1, help="Slot number, default: last used")
@click.option('--bip178', '-8', is_flag=True, 
                help="Use binary encoding defined in BIP-178")
@click.option('--bare', is_flag=True, 
                help="Remove text prefix proposed by Electrum")
@click.argument('cvc', type=str, metavar="[6-digit code]", required=False)
def dump_wif(cvc, slot, bip178, bare):
    "[SC] Show WIF for last unsealed slot, or give slot number"
    card = get_card(only_satscard=True)

    # guess most useful slot to show
    if slot == -1:
        st = card.get_status()
        active = st['slots'][0]
        slot = active-1 if active >= 1 else 0

    if bip178:
        bare = True

    cvc = cleanup_cvc(card, cvc)
    pk = card.get_privkey(cvc, slot)
    wif = render_wif(pk, bip_178=bip178, electrum=(not bare), testnet=card.is_testnet)

    if not bare:
        dump_key_info(slot, pk, wif, is_testnet=card.is_testnet)
    else:
        click.echo(wif)
    

def dump_key_info(slot_num, privkey, wif=None, is_testnet=False):
    # Show the WIF and address

    addr = render_address(privkey, is_testnet)

    wif = wif or render_wif(privkey, is_testnet)

    click.echo(f"Slot #{slot_num}:\n\n{addr}\n\n{wif}")

@main.command('setup')
@click.option('--chain-code', '-c', type=str, metavar="HEX",
                help="Chain code to be used for XPRV of resulting slot (32 bytes)")
@click.option('--new-chain-code', '-n', is_flag=True, 
                help="Pick a fresh chain code randomly [TS=default].")
@click.argument('cvc', type=str, metavar="(6-digit # code)", required=False)
def setup_slot(cvc, chain_code, new_chain_code):
    "Setup with a fresh private key."

    card = get_card()

    if card.is_tapsigner:
        # simpler?
        target = 0
        if not chain_code:
            new_chain_code = True
    else:
        # SATSCARD
        # only one possible value for slot number
        target = card.active_slot

        # but that slot must be un-used.
        resp = card.send('dump', slot=target)

        if resp.get('used', True):
            fail(f"Slot {target} has been used already. Unseal it, and move to next")

    args = dict(slot=target)

    if chain_code and new_chain_code:
        fail("Provide a chain code or make me pick one, not both")

    if new_chain_code:
        args['chain_code'] = sha256s(sha256s(os.urandom(128)))
    elif chain_code:
        try:
            chain_code = b2a_hex(chain_code)
            assert len(chain_code) == 32
        except:
            fail("Need 64 hex digits (32 bytes) for chain code.")
        args['chain_code'] = chain_code
    elif target == 0:
        # not expected case since factory setup on slot zero
        fail("Chain code required for slot zero setup")
    
    cvc = cleanup_cvc(card, cvc)
    ses_key, resp = card.send_auth('new', cvc, **args)

    if card.is_tapsigner:
        click.echo('TAPSIGNER ready for use')
    else:
        # only one field: new slot number
        card.active_slot = resp['slot']

        click.echo(card.address())

@main.command('balance')
@click.argument('cvc', type=str, metavar="(6-digit code)", required=False)
def show_balance(cvc):
    "[SC] Show the balance held on all slots"
    from cktap.sweep import UTXOList

    cvc = cleanup_cvc(cvc, missing_ok=True)
    card = get_card(only_satscard=True)

    rv = []
    click.echo('%-42s | Balance' % 'Address')
    click.echo(('-'*42) + '-+-------------')

    for slot in range(card.active_slot+1):
        addr = card.address(slot=slot, faster=True)
        if addr:
            b = UTXOList(addr)
            b.fetch()
            bal = b.balance()
            click.echo(f'{addr:40} | {bal}')
            

@main.command('core')
@click.option('--pretty', '-p', is_flag=True, help="Pretty-print JSON")
@click.argument('cvc', type=str, metavar="(6-digit code)", required=False)
def export_to_core(cvc, pretty):
    "[SC] Show JSON needed to import private keys into Bitcoin Core"

    # see 
    # - <https://bitcoincore.org/en/doc/0.21.0/rpc/wallet/importmulti/>
    # - <https://github.com/bitcoin/bitcoin/blob/master/doc/descriptors.md>
    # - <https://github.com/bitcoin/bips/blob/master/bip-0380.mediawiki>
    # - <https://github.com/bitcoin/bips/blob/master/bip-0382.mediawiki>

    card = get_card(only_satscard=True)

    # CVC is optional, but they probably want it
    if not cvc:
        click.echo("Warning: Without the code, can only watch addresses from this card.", err=1)
    cvc = cleanup_cvc(card, cvc, missing_ok=True)

    shared = dict(timestamp=PROJECT_EPOC_TIME_T, descr=[], internal=False)
    if not cvc:
        shared['watchonly'] = True

    rv = []
    for slot in range(card.active_slot+1):
        session_key, here = card.send_auth('dump', cvc, slot=slot)

        if here.get('used', None) == False:
            continue

        pk = None
        addr = None
        if here.get('sealed', None) == True:
            if cvc: continue
            pubkey, addr = card.address(incl_pubkey=1)

        if 'privkey' in here:
            pk = xor_bytes(session_key, here['privkey'])
            addr = render_address(pk, card.is_testnet)

        line = dict()
        line.update(shared)

        if pk:
            w = render_descriptor(privkey=pk, testnet=card.is_testnet)
        elif addr:
            w = render_descriptor(address=addr)
            line['pubkeys'] = [B2A(pubkey)]
        else:
            continue

        line['desc'] = [w]
        rv.append(line)

    # TODO this doesn't work, but looks likely
    click.echo('importmulti \'%s\'' % json.dumps(rv, indent=(2 if pretty else None)))

    
@main.command('status')
def card_status():
    "Show a few things about status of card"
    # meant as a starting point? no auth
    # TODO: make more useful, and be default cmd?
    card = get_card()
    st = card.get_status()
    print("-- TAPSIGNER Card --" if st.get('tapsigner', False) else '-- SATSCARD --')

    print(f'Card ident: {card.card_ident}')
    print(f'Birth Height: {card.birth_height}')

    if card.auth_delay:
        print(f'Needs auth delay: {card.auth_delay} seconds')

    if card.is_tapsigner:
        print(f'Number of backups: {st["num_backups"]}')
        if 'path' in st:
            print(f'Current derivation: ' + path2str(st['path']))
        else:
            print(f'No key picked yet')
    else:
        print(f'Active Slot: {card.active_slot or "first"}')
        print(f'Address: {card.address()}')

@main.command('xpub')
@click.option('--master', '-m', is_flag=True, help="Show master XPUB instead of derived")
@click.option('--show-path', '-p', is_flag=True, help="Show path as well")
@click.option('--skip-checks', '-x', is_flag=True, help="Skip extra checking of pubkey")
@click.argument('cvc', type=str, metavar="(6-digit code)", required=False)
def get_xpub(master, cvc, show_path, skip_checks):
    "[TS] Show the xpub in use"
    # meant as a starting point? no auth
    # TODO: make more useful, and be default cmd?

    card = get_card(only_tapsigner=True)
    cvc = cleanup_cvc(card, cvc)

    if show_path and not master:
        click.echo(card.get_derivation() + '\n')

    xpub = card.get_xpub(cvc, master)

    if not skip_checks:
        # extra checking; robust against MitM (cost = 20ms)
        from base58 import b58decode_check
        expect_pubkey = b58decode_check(xpub)[-33:]

        got_pubkey = card.get_pubkey(cvc)
        assert expect_pubkey == got_pubkey

    click.echo(xpub)

@main.command('json')
@click.argument('cvc', type=str, metavar="(6-digit code)", required=False)
def json_dump(cvc):
    "[TS] Dump wallet values in JSON format similar to Coldcard export."

    # - assumes you want specific BIP based on path observed
    # - change derivation before calling this if not what you want

    card = get_card(only_tapsigner=True)
    cvc = cleanup_cvc(card, cvc)

    path = card.get_derivation()
    assert path.startswith('m')
    path_comps = path.split('/')[1:]

    xfp = card.get_xfp(cvc).hex().upper()
    root_xpub = card.get_xpub(cvc, True)
    derived_xpub = card.get_xpub(cvc, False)

    # Mostly compat with Coldcard generic wallet export, but only one XPUB
    # see: coldcard/firmware/shared/export.py in generate_generic_export()

    rv = dict(chain='BTC' if not card.is_testnet else 'XTN',
                xpub = root_xpub,
                xfp = xfp,
                card_ident = card.card_ident,
            )

    if len(path_comps) == 3:
        rv['account'] = int(path_comps[-1][:-1])

    if len(path_comps) >= 2:
        # assuming BIP-44 compliance here
        bip_num = int(path_comps[0][:-1])
        if bip_num in { 44, 49, 84 }:
            rv[f'bip_{bip_num}'] = dict(deriv=path, xpub=derived_xpub,
                                        xfp=xfp, name=f'BIP-{bip_num}')

    print(json.dumps(rv, indent=2))

@main.command('backup')
@click.option('--outfile', '-o', metavar="filename.aes",
                        help="File to save into", default=None, type=str)
@click.option('--wrap-shell', '-s', is_flag=True, help="Wrap into shell script")
@click.argument('cvc', type=str, metavar="(6-digit code)", required=False)
def do_backup(cvc, outfile, wrap_shell):
    "[TS] Backup private key from card into AES-128-CTR encrypted file"
    card = get_card(only_tapsigner=True)
    cvc = cleanup_cvc(card, cvc)

    import datetime
    from base64 import b64encode

    if not outfile:
        nowish = datetime.datetime.now().isoformat()[0:16].replace(':', '')
        ident = card.card_ident.split('-')[0]
        outfile = f'backup-{ident}-{nowish}.' + ('sh' if wrap_shell else 'aes')

    enc = card.make_backup(cvc)

    with open(outfile, 'wt' if wrap_shell else 'wb') as fp:
        if not wrap_shell:
            fp.write(enc)
        else:
            b6 = b64encode(enc).decode('ascii')
            print(f'''\
#!/bin/sh
read -p "Key from back of card: " KEY
base64 -d <<END-OF-DATA | openssl aes-128-ctr -d -iv 0 -K $KEY | hexdump -C
{b6}
END-OF-DATA
echo "Look for 'chain_code' and 'priv_key' above."
echo "It's CBOR encoded data, visit https://cbor.io to decode."
''', file=fp)

        print(f"Wrote {fp.tell()} bytes to: {fp.name}")

@main.command('change')
@click.argument('cvc', type=str, metavar="(existing code)", required=False)
@click.argument('new_cvc', type=str, metavar="(new code)", required=False)
def change_cvc(cvc, new_cvc):
    "[TS] Change the CVC code (PIN code)"
    card = get_card(only_tapsigner=True)

    cvc = cleanup_cvc(card, cvc, prompt='existing code')
    new_cvc = cleanup_cvc(card, new_cvc, prompt='new code', confirm=True)

    card.change_cvc(cvc, new_cvc)

    click.echo("New code in effect now.")


@main.command('unlock')
def do_unlock():
    "Clear login delay (takes 15 seconds)"
    card = get_card()
    if card.auth_delay:
        with click.progressbar(label="Security delay", length=card.auth_delay) as bar:
            for n in range(card.auth_delay):
                card.send('wait')
                bar.update(1)

# EOF
